import numpy as np
import pytensor
import pytest
import statsmodels.api as sm

from numpy.testing import assert_allclose
from pytensor.graph.basic import explicit_graph_inputs
from scipy import linalg

from pymc_extras.statespace.models.ETS import BayesianETS
from pymc_extras.statespace.utils.constants import LONG_MATRIX_NAMES
from tests.statespace.shared_fixtures import rng
from tests.statespace.test_utilities import load_nile_test_data


@pytest.fixture(scope="session")
def data():
    return load_nile_test_data()


def test_invalid_order_raises():
    # Order must be length 3
    with pytest.raises(ValueError, match="Order must be a tuple of three strings"):
        BayesianETS(order=("A", "N"))

        # Order must be strings
    with pytest.raises(ValueError, match="Order must be a tuple of three strings"):
        BayesianETS(order=(2, 1, 1))

    # Only additive errors allowed
    with pytest.raises(ValueError, match="Only additive errors are supported"):
        BayesianETS(order=("M", "N", "N"))

    # Trend must be A or Ad
    with pytest.raises(ValueError, match="Invalid trend specification"):
        BayesianETS(order=("A", "P", "N"))

    # Seasonal must be A or N
    with pytest.raises(ValueError, match="Invalid seasonal specification"):
        BayesianETS(order=("A", "Ad", "M"))

    # seasonal_periods must be provided if seasonal is requested
    with pytest.raises(ValueError, match="If seasonal is True, seasonal_periods must be provided."):
        BayesianETS(order=("A", "Ad", "A"))


orders = (
    ("A", "N", "N"),
    ("A", "A", "N"),
    ("A", "Ad", "N"),
    ("A", "N", "A"),
    ("A", "A", "A"),
    ("A", "Ad", "A"),
)
order_names = (
    "Basic",
    "Trend",
    "Damped Trend",
    "Seasonal",
    "Trend and Seasonal",
    "Trend, Damped Trend, Seasonal",
)

order_expected_flags = (
    {"trend": False, "damped_trend": False, "seasonal": False},
    {"trend": True, "damped_trend": False, "seasonal": False},
    {"trend": True, "damped_trend": True, "seasonal": False},
    {"trend": False, "damped_trend": False, "seasonal": True},
    {"trend": True, "damped_trend": False, "seasonal": True},
    {"trend": True, "damped_trend": True, "seasonal": True},
)

order_params = (
    ["alpha", "initial_level"],
    ["alpha", "initial_level", "beta", "initial_trend"],
    ["alpha", "initial_level", "beta", "initial_trend", "phi"],
    ["alpha", "initial_level", "gamma", "initial_seasonal"],
    ["alpha", "initial_level", "beta", "initial_trend", "gamma", "initial_seasonal"],
    ["alpha", "initial_level", "beta", "initial_trend", "gamma", "initial_seasonal", "phi"],
)


@pytest.mark.parametrize(
    "order, expected_flags", zip(orders, order_expected_flags), ids=order_names
)
def test_order_flags(order, expected_flags):
    mod = BayesianETS(order=order, seasonal_periods=4)
    for key, value in expected_flags.items():
        assert getattr(mod, key) == value


def test_mode_argument():
    # Mode argument should be passed to the parent class
    mod = BayesianETS(order=("A", "N", "N"), mode="FAST_RUN")
    assert mod.mode == "FAST_RUN"


@pytest.mark.parametrize("order, expected_params", zip(orders, order_params), ids=order_names)
def test_param_info(order: tuple[str, str, str], expected_params):
    mod = BayesianETS(order=order, seasonal_periods=4)

    all_expected_params = [*expected_params, "sigma_state", "P0"]
    assert all(param in mod.param_names for param in all_expected_params)
    assert all(param in all_expected_params for param in mod.param_names)
    assert all(
        mod.param_info[param]["dims"] is None
        for param in expected_params
        if "seasonal" not in param
    )


@pytest.mark.parametrize("order, expected_params", zip(orders, order_params), ids=order_names)
@pytest.mark.parametrize("use_transformed", [True, False], ids=["transformed", "untransformed"])
def test_statespace_matrices(
    rng, order: tuple[str, str, str], expected_params: list[str], use_transformed: bool
):
    seasonal_periods = np.random.randint(3, 12)
    mod = BayesianETS(
        order=order,
        seasonal_periods=seasonal_periods,
        measurement_error=True,
        use_transformed_parameterization=use_transformed,
    )
    expected_states = 2 + int(order[1] != "N") + int(order[2] != "N") * seasonal_periods

    test_values = {
        "alpha": rng.beta(1, 1),
        "beta": rng.beta(1, 1),
        "gamma": rng.beta(1, 1),
        "phi": rng.beta(1, 1),
        "sigma_state": rng.normal() ** 2,
        "sigma_obs": rng.normal() ** 2,
        "initial_level": rng.normal() ** 2,
        "initial_trend": rng.normal() ** 2,
        "initial_seasonal": np.ones(seasonal_periods),
        "initial_state_cov": np.eye(expected_states),
    }

    matrices = x0, P0, c, d, T, Z, R, H, Q = mod._unpack_statespace_with_placeholders()

    assert x0.type.shape == (expected_states,)
    assert P0.type.shape == (expected_states, expected_states)
    assert c.type.shape == (expected_states,)
    assert d.type.shape == (1,)
    assert T.type.shape == (expected_states, expected_states)
    assert Z.type.shape == (1, expected_states)
    assert R.type.shape == (expected_states, 1)
    assert H.type.shape == (1, 1)
    assert Q.type.shape == (1, 1)

    inputs = list(explicit_graph_inputs(matrices))
    input_names = [x.name for x in inputs]
    assert all(name in input_names for name in expected_params)

    f_matrices = pytensor.function(inputs, matrices)
    [x0, P0, c, d, T, Z, R, H, Q] = f_matrices(**{name: test_values[name] for name in input_names})

    assert_allclose(H, np.eye(1) * test_values["sigma_obs"] ** 2)
    assert_allclose(Q, np.eye(1) * test_values["sigma_state"] ** 2)

    R_val = np.zeros((expected_states, 1))
    R_val[0] = 1.0 - test_values["alpha"]
    R_val[1] = test_values["alpha"]

    Z_val = np.zeros((1, expected_states))
    Z_val[0, 0] = 1.0
    Z_val[0, 1] = 1.0

    x0_val = np.zeros((expected_states,))
    x0_val[1] = test_values["initial_level"]

    if order[1] == "N":
        T_val = np.array([[0.0, 0.0], [0.0, 1.0]])
    else:
        x0_val[2] = test_values["initial_trend"]
        R_val[2] = (
            test_values["beta"] if use_transformed else test_values["beta"] * test_values["alpha"]
        )
        T_val = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 1.0], [0.0, 0.0, 1.0]])

    if order[1] == "Ad":
        T_val[1:, -1] *= test_values["phi"]

    if order[2] == "A":
        x0_val[2 + int(order[1] != "N") :] = test_values["initial_seasonal"]
        gamma = (
            test_values["gamma"]
            if use_transformed
            else (1 - test_values["alpha"]) * test_values["gamma"]
        )
        R_val[2 + int(order[1] != "N")] = gamma
        R_val[0] = R_val[0] - gamma

        S = np.eye(seasonal_periods, k=-1)
        S[0, -1] = 1.0
        Z_val[0, 2 + int(order[1] != "N")] = 1.0
    else:
        S = np.eye(0)

    T_val = linalg.block_diag(T_val, S)

    assert_allclose(x0, x0_val)
    assert_allclose(T, T_val)
    assert_allclose(R, R_val)
    assert_allclose(Z, Z_val)


@pytest.mark.parametrize("order, params", zip(orders, order_params), ids=order_names)
def test_statespace_matches_statsmodels(rng, order: tuple[str, str, str], params):
    seasonal_periods = rng.integers(3, 12)
    data = rng.normal(size=(100,))
    mod = BayesianETS(
        order=order,
        seasonal_periods=seasonal_periods,
        measurement_error=False,
        use_transformed_parameterization=True,
    )
    sm_mod = sm.tsa.statespace.ExponentialSmoothing(
        data,
        trend=mod.trend,
        damped_trend=mod.damped_trend,
        seasonal=seasonal_periods if mod.seasonal else None,
    )

    simplex_params = ["alpha", "beta", "gamma"]
    test_values = dict(zip(simplex_params, rng.dirichlet(alpha=np.ones(3))))
    test_values["phi"] = rng.beta(1, 1)

    test_values["initial_level"] = rng.normal()
    test_values["initial_trend"] = rng.normal()
    test_values["initial_seasonal"] = rng.normal(size=seasonal_periods)
    test_values["initial_state_cov"] = np.eye(mod.k_states)
    test_values["sigma_state"] = 1.0

    sm_test_values = test_values.copy()
    sm_test_values["smoothing_level"] = test_values["alpha"]
    sm_test_values["smoothing_trend"] = test_values["beta"]
    sm_test_values["smoothing_seasonal"] = test_values["gamma"]
    sm_test_values["damping_trend"] = test_values["phi"]
    sm_test_values["initial_seasonal"] = test_values["initial_seasonal"][0]
    for i in range(1, seasonal_periods):
        sm_test_values[f"initial_seasonal.L{i}"] = test_values["initial_seasonal"][i]

    vals = [
        np.atleast_1d(test_values[name])
        for name in ["initial_level", "initial_trend", "initial_seasonal"]
    ]
    x0 = np.concatenate([[0.0], *vals])

    mask = [True, True, order[1] != "N", *(order[2] != "N",) * seasonal_periods]

    sm_mod.initialize_known(initial_state=x0[mask], initial_state_cov=np.eye(mod.k_states))
    sm_mod.fit_constrained({name: sm_test_values[name] for name in sm_mod.param_names})

    matrices = mod._unpack_statespace_with_placeholders()
    inputs = list(explicit_graph_inputs(matrices))
    input_names = [x.name for x in inputs]

    f_matrices = pytensor.function(inputs, matrices)
    test_values_subset = {name: test_values[name] for name in input_names}

    matrices = f_matrices(**test_values_subset)
    sm_matrices = [sm_mod.ssm[name] for name in LONG_MATRIX_NAMES[2:]]

    for matrix, sm_matrix, name in zip(matrices[2:], sm_matrices, LONG_MATRIX_NAMES[2:]):
        assert_allclose(matrix, sm_matrix, err_msg=f"{name} does not match")


@pytest.mark.parametrize("order, params", zip(orders, order_params), ids=order_names)
@pytest.mark.parametrize("dense_cov", [True, False], ids=["dense", "diagonal"])
def test_ETS_with_multiple_endog(rng, order, params, dense_cov):
    seasonal_periods = 4
    mod = BayesianETS(
        order=order,
        seasonal_periods=seasonal_periods,
        measurement_error=False,
        use_transformed_parameterization=True,
        dense_innovation_covariance=dense_cov,
        endog_names=["A", "B"],
    )

    single_mod = BayesianETS(
        order=order,
        seasonal_periods=seasonal_periods,
        measurement_error=False,
        use_transformed_parameterization=True,
    )

    simplex_params = ["alpha", "beta", "gamma"]
    test_values = dict(zip(simplex_params, rng.dirichlet(alpha=np.ones(3), size=(mod.k_endog,)).T))
    test_values["phi"] = rng.beta(1, 1, size=(mod.k_endog,))

    test_values["initial_level"] = rng.normal(
        size=mod.k_endog,
    )
    test_values["initial_trend"] = rng.normal(
        size=mod.k_endog,
    )
    test_values["initial_seasonal"] = rng.normal(size=(mod.k_endog, seasonal_periods))
    test_values["initial_state_cov"] = np.eye(mod.k_states)

    if not dense_cov:
        test_values["sigma_state"] = np.ones(
            mod.k_endog,
        )
    else:
        L = np.random.normal(size=(mod.k_endog, mod.k_endog))
        test_values["state_cov"] = L @ L.T

    # Compile functions for the joined model
    matrices_pt = mod._unpack_statespace_with_placeholders()
    inputs = list(explicit_graph_inputs(matrices_pt))
    input_names = [x.name for x in inputs]

    test_values_subset = {name: test_values[name] for name in input_names}
    f_matrices = pytensor.function(inputs, matrices_pt)

    matrices = f_matrices(**test_values_subset)

    # Compile functions for the single model
    single_matrices_pt = single_mod._unpack_statespace_with_placeholders()
    single_inputs = list(explicit_graph_inputs(single_matrices_pt))
    single_input_names = [x.name for x in single_inputs]

    cursor = 0
    single_test_values_subsets = []
    for i in range(mod.k_endog):
        single_slice = slice(cursor, cursor + single_mod.k_states)
        d = {
            name: (
                test_values[name][i]
                if name != "initial_state_cov"
                else test_values_subset[name][single_slice, single_slice]
            )
            for name in single_input_names
            if name != "sigma_state"
        }
        if dense_cov:
            d["sigma_state"] = np.sqrt(test_values["state_cov"][i, i])
        else:
            d["sigma_state"] = test_values["sigma_state"][i]
        single_test_values_subsets.append(d)
        cursor += single_mod.k_states

    f_single_matrices = pytensor.function(single_inputs, single_matrices_pt)
    single_matrices = [f_single_matrices(**d) for d in single_test_values_subsets]
    names = [x.name for x in matrices_pt]

    for i, (x1, name) in enumerate(zip(matrices, names)):
        cursor = 0
        for j in range(mod.k_endog):
            x2 = single_matrices[j][i]
            state_slice = slice(cursor, cursor + single_mod.k_states)
            obs_slice = slice(j, j + 1)  # Also endog_slice -- it's doing double duty
            if name in ["state_intercept", "initial_state"]:
                assert_allclose(x1[state_slice], x2, err_msg=f"{name} does not match for case {j}")
            elif name in ["P0", "initial_state_cov", "transition"]:
                assert_allclose(
                    x1[state_slice, state_slice], x2, err_msg=f"{name} does not match for case {j}"
                )
            elif name == "selection":
                assert_allclose(
                    x1[state_slice, obs_slice], x2, err_msg=f"{name} does not match for case {j}"
                )
            elif name == "design":
                assert_allclose(
                    x1[obs_slice, state_slice], x2, err_msg=f"{name} does not match for case {j}"
                )
            elif name == "obs_intercept":
                assert_allclose(x1[obs_slice], x2, err_msg=f"{name} does not match for case {j}")
            elif name in ["obs_cov", "state_cov"]:
                assert_allclose(
                    x1[obs_slice, obs_slice], x2, err_msg=f"{name} does not match for case {j}"
                )
            else:
                raise ValueError(f"You forgot {name} !")

            cursor += single_mod.k_states


def test_ETS_stationary_initialization():
    mod = BayesianETS(
        order=("A", "Ad", "A"),
        seasonal_periods=4,
        stationary_initialization=True,
        initialization_dampening=0.66,
    )

    matrices = mod._unpack_statespace_with_placeholders()
    inputs = list(explicit_graph_inputs(matrices))
    input_names = [x.name for x in inputs]

    # Make sure the stationary_dampening dummy variables was completely rewritten away
    assert "stationary_dampening" not in input_names

    # P0 should have been removed from param names
    assert "P0" not in mod.param_names
    assert "P0" not in mod.param_info.keys()

    f = pytensor.function(inputs, matrices, mode="FAST_COMPILE")
    test_values = f(**{x.name: np.full(x.type.shape, 0.5) for x in inputs})
    outputs = {name: val for name, val in zip(LONG_MATRIX_NAMES, test_values)}

    # Make sure that the transition matrix has ones in the expected positions, not the model dampening factor
    assert outputs["transition"][1, 1] == 1.0
    assert outputs["transition"][2, 2] == 0.5  # phi = 0.5 -- trend is dampened anyway
    assert outputs["transition"][3, -1] == 1.0

    # P0 should be equal to the solution to the Lyapunov equation using the dampening factors in the transition matrix
    T_stationary = outputs["transition"].copy()
    T_stationary[1, 1] = mod.initialization_dampening
    T_stationary[3, -1] = mod.initialization_dampening

    R, Q = outputs["selection"], outputs["state_cov"]
    P0_expected = linalg.solve_discrete_lyapunov(T_stationary, R @ Q @ R.T)

    assert_allclose(outputs["initial_state_cov"], P0_expected, rtol=1e-8, atol=1e-8)



#   Copyright 2023 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import hashlib
import json
import sys
import tempfile

import numpy as np
import pandas as pd
import pymc as pm
import pytest

from pymc_extras.model_builder import ModelBuilder


@pytest.fixture(scope="module")
def toy_X():
    x = np.linspace(start=0, stop=1, num=100)
    X = pd.DataFrame({"input": x})
    return X


@pytest.fixture(scope="module")
def toy_y(toy_X):
    y = 5 * toy_X["input"] + 3
    y = y + np.random.normal(0, 1, size=len(toy_X))
    y = pd.Series(y, name="output")
    return y


def get_unfitted_model_instance(X, y):
    """Creates an unfitted model instance to which idata can be copied in
    and then used as a fitted model instance. That way a fitted model
    can be used multiple times without having to run `fit` multiple times."""
    sampler_config = {
        "draws": 20,
        "tune": 10,
        "chains": 2,
        "target_accept": 0.95,
    }
    model_config = {
        "a": {"loc": 0, "scale": 10, "dims": ("numbers",)},
        "b": {"loc": 0, "scale": 10},
        "obs_error": 2,
    }
    model = test_ModelBuilder(
        model_config=model_config, sampler_config=sampler_config, test_parameter="test_paramter"
    )
    # Do the things that `model.fit` does except sample to create idata.
    model._generate_and_preprocess_model_data(X, y.values.flatten())
    model.build_model(X, y)
    return model


@pytest.fixture(scope="module")
def fitted_model_instance_base(toy_X, toy_y):
    """Because fitting takes a relatively long time, this is intended to
    be used only once and then have new instances created and fit data patched in
    for tests that use a fitted model instance. Tests should use
    `fitted_model_instance` instead of this."""
    model = get_unfitted_model_instance(toy_X, toy_y)
    model.fit(toy_X, toy_y)
    return model


@pytest.fixture
def fitted_model_instance(toy_X, toy_y, fitted_model_instance_base):
    """Get a fitted model instance. A new instance is created and fit data is
    patched in, so tests using this fixture can modify the model object without
    affecting other tests."""
    model = get_unfitted_model_instance(toy_X, toy_y)
    model.idata = fitted_model_instance_base.idata.copy()
    return model


class test_ModelBuilder(ModelBuilder):
    def __init__(self, model_config=None, sampler_config=None, test_parameter=None):
        self.test_parameter = test_parameter
        super().__init__(model_config=model_config, sampler_config=sampler_config)

    _model_type = "test_model"
    version = "0.1"

    def build_model(self, X: pd.DataFrame, y: pd.Series, model_config=None):
        coords = {"numbers": np.arange(len(X))}
        self.generate_and_preprocess_model_data(X, y)
        with pm.Model(coords=coords) as self.model:
            if model_config is None:
                model_config = self.model_config
            x = pm.Data("x", self.X["input"].values)
            y_data = pm.Data("y_data", self.y)

            # prior parameters
            a_loc = model_config["a"]["loc"]
            a_scale = model_config["a"]["scale"]
            b_loc = model_config["b"]["loc"]
            b_scale = model_config["b"]["scale"]
            obs_error = model_config["obs_error"]

            # priors
            a = pm.Normal("a", a_loc, sigma=a_scale, dims=model_config["a"]["dims"])
            b = pm.Normal("b", b_loc, sigma=b_scale)
            obs_error = pm.HalfNormal("σ_model_fmc", obs_error)

            # observed data
            output = pm.Normal("output", a + b * x, obs_error, shape=x.shape, observed=y_data)

    def _save_input_params(self, idata):
        idata.attrs["test_paramter"] = json.dumps(self.test_parameter)

    @property
    def output_var(self):
        return "output"

    def _data_setter(self, x: pd.Series, y: pd.Series = None):
        with self.model:
            pm.set_data({"x": x.values})
            if y is not None:
                pm.set_data({"y_data": y.values})

    @property
    def _serializable_model_config(self):
        return self.model_config

    def generate_and_preprocess_model_data(self, X: pd.DataFrame, y: pd.Series):
        self.X = X
        self.y = y

    @staticmethod
    def get_default_model_config() -> dict:
        return {
            "a": {"loc": 0, "scale": 10, "dims": ("numbers",)},
            "b": {"loc": 0, "scale": 10},
            "obs_error": 2,
        }

    def _generate_and_preprocess_model_data(
        self, X: pd.DataFrame | pd.Series, y: pd.Series
    ) -> None:
        self.X = X
        self.y = y

    @staticmethod
    def get_default_sampler_config() -> dict:
        return {
            "draws": 10,
            "tune": 10,
            "chains": 3,
            "target_accept": 0.95,
        }


def test_save_input_params(fitted_model_instance):
    assert fitted_model_instance.idata.attrs["test_paramter"] == '"test_paramter"'


@pytest.mark.skipif(
    sys.platform == "win32", reason="Permissions for temp files not granted on windows CI."
)
def test_save_load(fitted_model_instance):
    temp = tempfile.NamedTemporaryFile(mode="w", encoding="utf-8", delete=False)
    fitted_model_instance.save(temp.name)
    test_builder2 = test_ModelBuilder.load(temp.name)
    assert fitted_model_instance.idata.groups() == test_builder2.idata.groups()
    assert fitted_model_instance.id == test_builder2.id
    x_pred = np.random.uniform(low=0, high=1, size=100)
    prediction_data = pd.DataFrame({"input": x_pred})
    pred1 = fitted_model_instance.predict(prediction_data["input"])
    pred2 = test_builder2.predict(prediction_data["input"])
    assert pred1.shape == pred2.shape
    temp.close()


def test_initial_build_and_fit(fitted_model_instance, check_idata=True) -> ModelBuilder:
    if check_idata:
        assert fitted_model_instance.idata is not None
        assert "posterior" in fitted_model_instance.idata.groups()


def test_save_without_fit_raises_runtime_error():
    model_builder = test_ModelBuilder()
    with pytest.raises(RuntimeError):
        model_builder.save("saved_model")


def test_empty_sampler_config_fit(toy_X, toy_y):
    sampler_config = {}
    model_builder = test_ModelBuilder(sampler_config=sampler_config)
    model_builder.idata = model_builder.fit(X=toy_X, y=toy_y)
    assert model_builder.idata is not None
    assert "posterior" in model_builder.idata.groups()


def test_fit(fitted_model_instance):
    prediction_data = pd.DataFrame({"input": np.random.uniform(low=0, high=1, size=100)})
    pred = fitted_model_instance.predict(prediction_data["input"])
    post_pred = fitted_model_instance.sample_posterior_predictive(
        prediction_data["input"], extend_idata=True, combined=True
    )
    post_pred[fitted_model_instance.output_var].shape[0] == prediction_data.input.shape


def test_fit_no_y(toy_X):
    model_builder = test_ModelBuilder()
    model_builder.idata = model_builder.fit(X=toy_X, chains=1, tune=1, draws=1)
    assert model_builder.model is not None
    assert model_builder.idata is not None
    assert "posterior" in model_builder.idata.groups()


def test_predict(fitted_model_instance):
    x_pred = np.random.uniform(low=0, high=1, size=100)
    prediction_data = pd.DataFrame({"input": x_pred})
    pred = fitted_model_instance.predict(prediction_data["input"])
    # Perform elementwise comparison using numpy
    assert isinstance(pred, np.ndarray)
    assert len(pred) > 0


@pytest.mark.parametrize("combined", [True, False])
def test_sample_posterior_predictive(fitted_model_instance, combined):
    n_pred = 100
    x_pred = np.random.uniform(low=0, high=1, size=n_pred)
    prediction_data = pd.DataFrame({"input": x_pred})
    pred = fitted_model_instance.sample_posterior_predictive(
        prediction_data["input"], combined=combined, extend_idata=True
    )
    chains = fitted_model_instance.idata.sample_stats.sizes["chain"]
    draws = fitted_model_instance.idata.sample_stats.sizes["draw"]
    expected_shape = (n_pred, chains * draws) if combined else (chains, draws, n_pred)
    assert pred[fitted_model_instance.output_var].shape == expected_shape
    assert np.issubdtype(pred[fitted_model_instance.output_var].dtype, np.floating)


@pytest.mark.parametrize("group", ["prior_predictive", "posterior_predictive"])
@pytest.mark.parametrize("extend_idata", [True, False])
def test_sample_xxx_extend_idata_param(fitted_model_instance, group, extend_idata):
    output_var = fitted_model_instance.output_var
    idata_prev = fitted_model_instance.idata[group][output_var]

    # Since coordinates are provided, the dimension must match
    n_pred = 100  # Must match toy_x
    x_pred = np.random.uniform(0, 1, n_pred)

    prediction_data = pd.DataFrame({"input": x_pred})
    if group == "prior_predictive":
        prediction_method = fitted_model_instance.sample_prior_predictive
    else:  # group == "posterior_predictive":
        prediction_method = fitted_model_instance.sample_posterior_predictive

    pred = prediction_method(prediction_data["input"], combined=False, extend_idata=extend_idata)

    pred_unstacked = pred[output_var].values
    idata_now = fitted_model_instance.idata[group][output_var].values

    if extend_idata:
        # After sampling, data in the model should be the same as the predictions
        np.testing.assert_array_equal(idata_now, pred_unstacked)
        # Data in the model should NOT be the same as before
        if idata_now.shape == idata_prev.values.shape:
            assert np.sum(np.abs(idata_now - idata_prev.values) < 1e-5) <= 2
    else:
        # After sampling, data in the model should be the same as it was before
        np.testing.assert_array_equal(idata_now, idata_prev.values)
        # Data in the model should NOT be the same as the predictions
        if idata_now.shape == pred_unstacked.shape:
            assert np.sum(np.abs(idata_now - pred_unstacked) < 1e-5) <= 2


def test_model_config_formatting():
    model_config = {
        "a": {
            "loc": [0, 0],
            "scale": 10,
            "dims": [
                "x",
            ],
        },
    }
    model_builder = test_ModelBuilder()
    converted_model_config = model_builder._model_config_formatting(model_config)
    np.testing.assert_equal(converted_model_config["a"]["dims"], ("x",))
    np.testing.assert_equal(converted_model_config["a"]["loc"], np.array([0, 0]))


def test_id():
    model_builder = test_ModelBuilder()
    expected_id = hashlib.sha256(
        str(model_builder.model_config.values()).encode()
        + model_builder.version.encode()
        + model_builder._model_type.encode()
    ).hexdigest()[:16]

    assert model_builder.id == expected_id



import numpy as np
import pytest

from pytensor import tensor as pt

from pymc_extras.inference.laplace_approx import scipy_interface


@pytest.fixture
def simple_loss_and_inputs():
    x = pt.vector("x")
    loss = pt.sum(x**2)
    return loss, [x]


def test_compile_functions_for_scipy_optimize_loss_only(simple_loss_and_inputs):
    loss, inputs = simple_loss_and_inputs
    funcs = scipy_interface._compile_functions_for_scipy_optimize(
        loss, inputs, compute_grad=False, compute_hess=False, compute_hessp=False
    )
    assert len(funcs) == 1
    f_loss = funcs[0]
    x_val = np.array([1.0, 2.0, 3.0])
    result = f_loss(x_val)
    assert np.isclose(result, np.sum(x_val**2))


def test_compile_functions_for_scipy_optimize_with_grad(simple_loss_and_inputs):
    loss, inputs = simple_loss_and_inputs
    funcs = scipy_interface._compile_functions_for_scipy_optimize(
        loss, inputs, compute_grad=True, compute_hess=False, compute_hessp=False
    )
    f_fused = funcs[0]
    x_val = np.array([1.0, 2.0, 3.0])
    loss_val, grad_val = f_fused(x_val)
    assert np.isclose(loss_val, np.sum(x_val**2))
    assert np.allclose(grad_val, 2 * x_val)


def test_compile_functions_for_scipy_optimize_with_hess(simple_loss_and_inputs):
    loss, inputs = simple_loss_and_inputs
    funcs = scipy_interface._compile_functions_for_scipy_optimize(
        loss, inputs, compute_grad=True, compute_hess=True, compute_hessp=False
    )
    f_fused = funcs[0]
    x_val = np.array([1.0, 2.0])
    loss_val, grad_val, hess_val = f_fused(x_val)
    assert np.isclose(loss_val, np.sum(x_val**2))
    assert np.allclose(grad_val, 2 * x_val)
    assert np.allclose(hess_val, 2 * np.eye(len(x_val)))


def test_compile_functions_for_scipy_optimize_with_hessp(simple_loss_and_inputs):
    loss, inputs = simple_loss_and_inputs
    funcs = scipy_interface._compile_functions_for_scipy_optimize(
        loss, inputs, compute_grad=True, compute_hess=False, compute_hessp=True
    )
    f_fused, f_hessp = funcs
    x_val = np.array([1.0, 2.0])
    p_val = np.array([1.0, 0.0])

    loss_val, grad_val = f_fused(x_val)
    assert np.isclose(loss_val, np.sum(x_val**2))
    assert np.allclose(grad_val, 2 * x_val)

    hessp_val = f_hessp(x_val, p_val)
    assert np.allclose(hessp_val, 2 * p_val)


def test_scipy_optimize_funcs_from_loss_invalid_backend(simple_loss_and_inputs):
    loss, inputs = simple_loss_and_inputs
    with pytest.raises(ValueError, match="Invalid gradient backend"):
        scipy_interface.scipy_optimize_funcs_from_loss(
            loss,
            inputs,
            {"x": np.array([1.0, 2.0])},
            use_grad=True,
            use_hess=False,
            use_hessp=False,
            gradient_backend="not_a_backend",
        )


def test_scipy_optimize_funcs_from_loss_hess_without_grad(simple_loss_and_inputs):
    loss, inputs = simple_loss_and_inputs
    with pytest.raises(
        ValueError, match="Cannot compute hessian without also computing the gradient"
    ):
        scipy_interface.scipy_optimize_funcs_from_loss(
            loss,
            inputs,
            {"x": np.array([1.0, 2.0])},
            use_grad=False,
            use_hess=True,
            use_hessp=False,
        )


@pytest.mark.parametrize("backend", ["pytensor", "jax"], ids=str)
def test_scipy_optimize_funcs_from_loss_backend(backend, simple_loss_and_inputs):
    if backend == "jax":
        pytest.importorskip("jax", reason="JAX is not installed")

    loss, inputs = simple_loss_and_inputs
    f_fused, f_hessp = scipy_interface.scipy_optimize_funcs_from_loss(
        loss,
        inputs,
        {"x": np.array([1.0, 2.0])},
        use_grad=True,
        use_hess=False,
        use_hessp=False,
        gradient_backend=backend,
    )
    x_val = np.array([1.0, 2.0])
    loss_val, grad_val = f_fused(x_val)
    assert np.isclose(loss_val, np.sum(x_val**2))
    assert np.allclose(grad_val, 2 * x_val)
    assert f_hessp is None



from pymc_extras.statespace.models.structural.components.autoregressive import (
    AutoregressiveComponent,
)
from pymc_extras.statespace.models.structural.components.cycle import CycleComponent
from pymc_extras.statespace.models.structural.components.level_trend import LevelTrendComponent
from pymc_extras.statespace.models.structural.components.measurement_error import MeasurementError
from pymc_extras.statespace.models.structural.components.regression import RegressionComponent
from pymc_extras.statespace.models.structural.components.seasonality import (
    FrequencySeasonality,
    TimeSeasonality,
)

__all__ = [
    "AutoregressiveComponent",
    "CycleComponent",
    "FrequencySeasonality",
    "LevelTrendComponent",
    "MeasurementError",
    "RegressionComponent",
    "TimeSeasonality",
]



import numpy as np

from pymc_extras.statespace.models.structural.core import Component


class MeasurementError(Component):
    r"""
    Measurement error component for structural time series models.

    This component adds observation noise to the model by introducing a variance parameter
    that affects the observation covariance matrix H. Unlike other components, it has no
    hidden states and should only be used in combination with other components.

    Parameters
    ----------
    name : str, optional
        Name of the measurement error component. Default is "MeasurementError".
    observed_state_names : list[str] | None, optional
        Names of the observed variables. If None, defaults to ["data"].
    share_states: bool, default False
        Whether latent states are shared across the observed states. If True, there will be only one set of latent
        states, which are observed by all observed states. If False, each observed state has its own set of
        latent states. This argument has no effect if `k_endog` is 1.

    Notes
    -----
    The measurement error component models observation noise as:

    .. math::

        y_t = \text{signal}_t + \varepsilon_t, \quad \varepsilon_t \sim N(0, \sigma^2)

    Where :math:`\text{signal}_t` is the true signal from other components and
    :math:`\sigma^2` is the measurement error variance.

    This component:
        - Has no hidden states (k_states = 0)
        - Has no innovations (k_posdef = 0)
        - Adds a single parameter: sigma_{name}
        - Modifies the observation covariance matrix H

    Examples
    --------
    **Basic usage with trend component:**

    .. code:: python

        from pymc_extras.statespace import structural as st
        import pymc as pm
        import pytensor.tensor as pt

        trend = st.LevelTrendComponent(order=2, innovations_order=1)
        error = st.MeasurementError()

        ss_mod = (trend + error).build()

        # Use with PyMC
        with pm.Model(coords=ss_mod.coords) as model:
            P0 = pm.Deterministic('P0', pt.eye(ss_mod.k_states) * 10, dims=ss_mod.param_dims['P0'])
            initial_trend = pm.Normal('initial_trend', sigma=10, dims=ss_mod.param_dims['initial_trend'])
            sigma_obs = pm.Exponential('sigma_obs', 1, dims=ss_mod.param_dims['sigma_obs'])

            ss_mod.build_statespace_graph(data)
            idata = pm.sample()

    **Multivariate measurement error:**

    .. code:: python

        # For multiple observed variables
        # This creates separate measurement error variances for each variable
        # sigma_obs_error will have shape (3,) for the three variables
        error = st.MeasurementError(
            name="obs_error",
            observed_state_names=["gdp", "unemployment", "inflation"]
        )

    **Complete model example:**

    .. code:: python

        trend = st.LevelTrendComponent(order=2, innovations_order=1)
        seasonal = st.TimeSeasonality(season_length=12, innovations=True)
        error = st.MeasurementError()

        model = (trend + seasonal + error).build()

        # The model now includes:
        # - Trend parameters: level_trend, sigma_trend
        # - Seasonal parameters: seasonal_coefs, sigma_seasonal
        # - Measurement error parameter: sigma_obs

    See Also
    --------
    Component : Base class for all structural components.
    StructuralTimeSeries : Complete model class.
    """

    def __init__(
        self,
        name: str = "MeasurementError",
        observed_state_names: list[str] | None = None,
        share_states: bool = False,
    ):
        if observed_state_names is None:
            observed_state_names = ["data"]

        self.share_states = share_states

        k_endog = len(observed_state_names)
        k_states = 0
        k_posdef = 0

        super().__init__(
            name,
            k_endog,
            k_states,
            k_posdef,
            measurement_error=True,
            combine_hidden_states=False,
            observed_state_names=observed_state_names,
            share_states=share_states,
        )

    def populate_component_properties(self):
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        self.param_names = [f"sigma_{self.name}"]
        self.param_dims = {}
        self.coords = {}

        if k_endog_effective > 1:
            self.param_dims[f"sigma_{self.name}"] = (f"endog_{self.name}",)
            self.coords[f"endog_{self.name}"] = self.observed_state_names

        self.param_info = {
            f"sigma_{self.name}": {
                "shape": (k_endog_effective,) if k_endog_effective > 1 else (),
                "constraints": "Positive",
                "dims": (f"endog_{self.name}",) if k_endog_effective > 1 else None,
            }
        }

    def make_symbolic_graph(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        sigma_shape = () if k_endog_effective == 1 else (k_endog_effective,)
        error_sigma = self.make_and_register_variable(f"sigma_{self.name}", shape=sigma_shape)

        diag_idx = np.diag_indices(self.k_endog)
        idx = np.s_["obs_cov", diag_idx[0], diag_idx[1]]
        self.ssm[idx] = error_sigma**2



import numpy as np
import pytensor.tensor as pt

from pymc_extras.statespace.models.structural.core import Component
from pymc_extras.statespace.models.structural.utils import order_to_mask
from pymc_extras.statespace.utils.constants import AR_PARAM_DIM


class AutoregressiveComponent(Component):
    r"""
    Autoregressive timeseries component

    Parameters
    ----------
    order: int or sequence of int

        If int, the number of lags to include in the model.
        If a sequence, an array-like of zeros and ones indicating which lags to include in the model.

    name: str, default "auto_regressive"
        A name for this autoregressive component. Used to label dimensions and coordinates.

    observed_state_names: list[str] | None, default None
        List of strings for observed state labels. If None, defaults to ["data"].

    share_states: bool, default False
        Whether latent states are shared across the observed states. If True, there will be only one set of latent
        states, which are observed by all observed states. If False, each observed state has its own set of
        latent states. This argument has no effect if `k_endog` is 1.

    Notes
    -----
    An autoregressive component can be thought of as a way o introducing serially correlated errors into the model.
    The process is modeled:

    .. math::
        x_t = \sum_{i=1}^p \rho_i x_{t-i}

    Where ``p``, the number of autoregressive terms to model, is the order of the process. By default, all lags up to
    ``p`` are included in the model. To disable lags, pass a list of zeros and ones to the ``order`` argumnet. For
    example, ``order=[1, 1, 0, 1]`` would become:

    .. math::
        x_t = \rho_1 x_{t-1} + \rho_2 x_{t-1} + \rho_4 x_{t-1}

    The coefficient :math:`\rho_3` has been constrained to zero.

    .. warning:: This class is meant to be used as a component in a structural time series model. For modeling of
              stationary processes with ARIMA, use ``statespace.BayesianSARIMAX``.

    Examples
    --------
    Model a timeseries as an AR(2) process with non-zero mean:

    .. code:: python

        from pymc_extras.statespace import structural as st
        import pymc as pm
        import pytensor.tensor as pt

        trend = st.LevelTrendComponent(order=1, innovations_order=0)
        ar = st.AutoregressiveComponent(2)
        ss_mod = (trend + ar).build()

        with pm.Model(coords=ss_mod.coords) as model:
            P0 = pm.Deterministic('P0', pt.eye(ss_mod.k_states) * 10, dims=ss_mod.param_dims['P0'])
            intitial_trend = pm.Normal('initial_trend', sigma=10, dims=ss_mod.param_dims['initial_trend'])
            ar_params = pm.Normal('ar_params', dims=ss_mod.param_dims['ar_params'])
            sigma_ar = pm.Exponential('sigma_ar', 1, dims=ss_mod.param_dims['sigma_ar'])

            ss_mod.build_statespace_graph(data)
            idata = pm.sample(nuts_sampler='numpyro')

    """

    def __init__(
        self,
        order: int = 1,
        name: str = "auto_regressive",
        observed_state_names: list[str] | None = None,
        share_states: bool = False,
    ):
        if observed_state_names is None:
            observed_state_names = ["data"]

        k_endog = len(observed_state_names)
        k_endog_effective = k_posdef = 1 if share_states else k_endog

        order = order_to_mask(order)
        ar_lags = np.flatnonzero(order).ravel().astype(int) + 1
        k_states = len(order)

        self.share_states = share_states
        self.order = order
        self.ar_lags = ar_lags

        super().__init__(
            name=name,
            k_endog=k_endog,
            k_states=k_states * k_endog_effective,
            k_posdef=k_posdef,
            measurement_error=True,
            combine_hidden_states=True,
            observed_state_names=observed_state_names,
            obs_state_idxs=np.tile(np.r_[[1.0], np.zeros(k_states - 1)], k_endog_effective),
            share_states=share_states,
        )

    def populate_component_properties(self):
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective  # this is also the number of AR lags
        base_names = [f"L{i + 1}_{self.name}" for i in range(k_states)]

        if self.share_states:
            self.state_names = [f"{name}[shared]" for name in base_names]
            self.shock_names = [f"{self.name}[shared]"]
        else:
            self.state_names = [
                f"{name}[{state_name}]"
                for state_name in self.observed_state_names
                for name in base_names
            ]
            self.shock_names = [
                f"{self.name}[{obs_name}]" for obs_name in self.observed_state_names
            ]

        self.param_names = [f"params_{self.name}", f"sigma_{self.name}"]
        self.param_dims = {f"params_{self.name}": (f"lag_{self.name}",)}
        self.coords = {f"lag_{self.name}": self.ar_lags.tolist()}

        if k_endog_effective > 1:
            self.param_dims[f"params_{self.name}"] = (
                f"endog_{self.name}",
                f"lag_{self.name}",
            )
            self.param_dims[f"sigma_{self.name}"] = (f"endog_{self.name}",)

            self.coords[f"endog_{self.name}"] = self.observed_state_names

        self.param_info = {
            f"params_{self.name}": {
                "shape": (k_states,) if self.k_endog == 1 else (self.k_endog, k_states),
                "constraints": None,
                "dims": (AR_PARAM_DIM,)
                if self.k_endog == 1
                else (
                    f"endog_{self.name}",
                    f"lag_{self.name}",
                ),
            },
            f"sigma_{self.name}": {
                "shape": () if self.k_endog == 1 else (self.k_endog,),
                "constraints": "Positive",
                "dims": None if self.k_endog == 1 else (f"endog_{self.name}",),
            },
        }

    def make_symbolic_graph(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective
        k_posdef = self.k_posdef

        k_nonzero = int(sum(self.order))
        ar_params = self.make_and_register_variable(
            f"params_{self.name}",
            shape=(k_nonzero,) if k_endog_effective == 1 else (k_endog_effective, k_nonzero),
        )
        sigma_ar = self.make_and_register_variable(
            f"sigma_{self.name}", shape=() if k_endog_effective == 1 else (k_endog_effective,)
        )

        if k_endog_effective == 1:
            T = pt.eye(k_states, k=-1)
            ar_idx = (np.zeros(k_nonzero, dtype="int"), np.nonzero(self.order)[0])
            T = T[ar_idx].set(ar_params)

        else:
            transition_matrices = []

            for i in range(k_endog_effective):
                T = pt.eye(k_states, k=-1)
                ar_idx = (np.zeros(k_nonzero, dtype="int"), np.nonzero(self.order)[0])
                T = T[ar_idx].set(ar_params[i])
                transition_matrices.append(T)
            T = pt.specify_shape(
                pt.linalg.block_diag(*transition_matrices), (self.k_states, self.k_states)
            )

        self.ssm["transition", :, :] = T

        R = np.eye(k_states)
        R_mask = np.full((k_states,), False)
        R_mask[0] = True
        R = R[:, R_mask]

        self.ssm["selection", :, :] = pt.specify_shape(
            pt.linalg.block_diag(*[R for _ in range(k_endog_effective)]), (self.k_states, k_posdef)
        )

        Zs = [pt.zeros((1, k_states))[0, 0].set(1.0) for _ in range(k_endog)]

        if self.share_states:
            Z = pt.join(0, *Zs)
        else:
            Z = pt.linalg.block_diag(*Zs)
        self.ssm["design", :, :] = pt.specify_shape(Z, (k_endog, self.k_states))

        cov_idx = ("state_cov", *np.diag_indices(k_posdef))
        self.ssm[cov_idx] = sigma_ar**2



import pytensor.tensor as pt

from pytensor.tensor.nlinalg import matrix_dot

from pymc_extras.statespace.utils.constants import JITTER_DEFAULT, NEVER_TIME_VARYING, VECTOR_VALUED


def decide_if_x_time_varies(x, name):
    if name in NEVER_TIME_VARYING:
        return False

    ndim = x.ndim

    if name in VECTOR_VALUED:
        if ndim not in [1, 2]:
            raise ValueError(
                f"Vector {name} has {ndim} dimensions; it should have either 1 (static),"
                f" or 2 (time varying )"
            )

        return ndim == 2

    if ndim not in [2, 3]:
        raise ValueError(
            f"Matrix {name} has {ndim} dimensions; it should have either"
            f" 2 (static), or 3 (time varying)."
        )

    return ndim == 3


def split_vars_into_seq_and_nonseq(params, param_names):
    """
    Split inputs into those that are time varying and those that are not. This division is required by scan.
    """
    sequences, non_sequences = [], []
    seq_names, non_seq_names = [], []

    for param, name in zip(params, param_names):
        if decide_if_x_time_varies(param, name):
            sequences.append(param)
            seq_names.append(name)
        else:
            non_sequences.append(param)
            non_seq_names.append(name)

    return sequences, non_sequences, seq_names, non_seq_names


def stabilize(cov, jitter=JITTER_DEFAULT):
    # Ensure diagonal is non-zero
    cov = cov + pt.identity_like(cov) * jitter

    return cov


def quad_form_sym(A, B):
    out = matrix_dot(A, B, A.T)
    return 0.5 * (out + out.T)



import numpy as np
import pytensor
import pytest

from numpy.testing import assert_allclose
from pytensor import config
from pytensor.graph.basic import explicit_graph_inputs

from pymc_extras.statespace.models import structural as st
from tests.statespace.models.structural.conftest import _assert_basic_coords_correct
from tests.statespace.test_utilities import simulate_from_numpy_model


@pytest.mark.parametrize("order", [1, 2, [1, 0, 1]], ids=["AR1", "AR2", "AR(1,0,1)"])
def test_autoregressive_model(order, rng):
    ar = st.AutoregressiveComponent(order=order).build(verbose=False)

    _assert_basic_coords_correct(ar)

    lags = np.arange(len(order) if isinstance(order, list) else order, dtype="int") + 1
    if isinstance(order, list):
        lags = lags[np.flatnonzero(order)]
    assert_allclose(ar.coords["lag_auto_regressive"], lags)


def test_autoregressive_multiple_observed_build(rng):
    ar = st.AutoregressiveComponent(order=3, name="ar", observed_state_names=["data_1", "data_2"])
    mod = ar.build(verbose=False)

    assert mod.k_endog == 2
    assert mod.k_states == 6
    assert mod.k_posdef == 2

    assert mod.state_names == [
        "L1_ar[data_1]",
        "L2_ar[data_1]",
        "L3_ar[data_1]",
        "L1_ar[data_2]",
        "L2_ar[data_2]",
        "L3_ar[data_2]",
    ]

    assert mod.shock_names == ["ar[data_1]", "ar[data_2]"]

    params = {
        "params_ar": np.full(
            (
                2,
                sum(ar.order),
            ),
            0.5,
            dtype=config.floatX,
        ),
        "sigma_ar": np.array([0.05, 0.12]),
    }
    _, _, _, _, T, Z, R, _, Q = mod._unpack_statespace_with_placeholders()
    input_vars = explicit_graph_inputs([T, Z, R, Q])
    fn = pytensor.function(
        inputs=list(input_vars),
        outputs=[T, Z, R, Q],
        mode="FAST_COMPILE",
    )

    T, Z, R, Q = fn(**params)

    np.testing.assert_allclose(
        T,
        np.array(
            [
                [0.5, 0.5, 0.5, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.5, 0.5, 0.5],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            ]
        ),
    )

    np.testing.assert_allclose(
        Z, np.array([[1.0, 0.0, 0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0, 0.0, 0.0]])
    )

    np.testing.assert_allclose(
        R, np.array([[1.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 1.0], [0.0, 0.0], [0.0, 0.0]])
    )

    np.testing.assert_allclose(Q, np.diag([0.05**2, 0.12**2]))


def test_autoregressive_multiple_observed_shared():
    ar = st.AutoregressiveComponent(
        order=1,
        name="latent",
        observed_state_names=["data_1", "data_2", "data_3"],
        share_states=True,
    )
    mod = ar.build(verbose=False)

    assert mod.k_endog == 3
    assert mod.k_states == 1
    assert mod.k_posdef == 1

    assert mod.state_names == ["L1_latent[shared]"]
    assert mod.shock_names == ["latent[shared]"]
    assert mod.coords["lag_latent"] == [1]
    assert "endog_latent" not in mod.coords

    outputs = [mod.ssm["transition"], mod.ssm["design"]]
    params = {"params_latent": np.array([0.9])}
    T, Z = pytensor.function(list(explicit_graph_inputs(outputs)), outputs)(**params)

    np.testing.assert_allclose(np.array([[1.0], [1.0], [1.0]]), Z)

    np.testing.assert_allclose(np.array([[0.9]]), T)


def test_autoregressive_multiple_observed_data(rng):
    ar = st.AutoregressiveComponent(order=1, observed_state_names=["data_1", "data_2", "data_3"])
    mod = ar.build(verbose=False)

    params = {
        "params_auto_regressive": np.array([0.9, 0.8, 0.5]).reshape((3, 1)),
        "sigma_auto_regressive": np.array([0.05, 0.12, 0.22]),
        "initial_state_cov": np.eye(3),
    }

    # Recover the AR(1) coefficients from the simulated data via OLS
    x, y = simulate_from_numpy_model(mod, rng, params, steps=2000)
    for i in range(3):
        ols_coefs = np.polyfit(y[:-1, i], y[1:, i], 1)
        np.testing.assert_allclose(ols_coefs[0], params["params_auto_regressive"][i, 0], atol=1e-1)


def test_add_autoregressive_different_observed():
    mod_1 = st.AutoregressiveComponent(order=1, name="ar1", observed_state_names=["data_1"])
    mod_2 = st.AutoregressiveComponent(name="ar6", order=6, observed_state_names=["data_2"])

    mod = (mod_1 + mod_2).build(verbose=False)

    assert mod.k_endog == 2
    assert mod.k_states == 7
    assert mod.k_posdef == 2
    assert mod.state_names == [
        f"L1_{mod_1.name}[data_1]",
        f"L1_{mod_2.name}[data_2]",
        f"L2_{mod_2.name}[data_2]",
        f"L3_{mod_2.name}[data_2]",
        f"L4_{mod_2.name}[data_2]",
        f"L5_{mod_2.name}[data_2]",
        f"L6_{mod_2.name}[data_2]",
    ]

    assert mod.shock_names == ["ar1[data_1]", "ar6[data_2]"]
    assert mod.coords["lag_ar1"] == [1]
    assert mod.coords["lag_ar6"] == [1, 2, 3, 4, 5, 6]


def test_autoregressive_shared_and_not_shared():
    shared = st.AutoregressiveComponent(
        order=3,
        name="shared_ar",
        observed_state_names=["data_1", "data_2", "data_3"],
        share_states=True,
    )
    individual = st.AutoregressiveComponent(
        order=3,
        name="individual_ar",
        observed_state_names=["data_1", "data_2", "data_3"],
        share_states=False,
    )

    mod = (shared + individual).build(verbose=False)

    assert mod.k_endog == 3
    assert mod.k_states == 3 + 3 * 3
    assert mod.k_posdef == 4

    assert mod.state_names == [
        "L1_shared_ar[shared]",
        "L2_shared_ar[shared]",
        "L3_shared_ar[shared]",
        "L1_individual_ar[data_1]",
        "L2_individual_ar[data_1]",
        "L3_individual_ar[data_1]",
        "L1_individual_ar[data_2]",
        "L2_individual_ar[data_2]",
        "L3_individual_ar[data_2]",
        "L1_individual_ar[data_3]",
        "L2_individual_ar[data_3]",
        "L3_individual_ar[data_3]",
    ]

    assert mod.shock_names == [
        "shared_ar[shared]",
        "individual_ar[data_1]",
        "individual_ar[data_2]",
        "individual_ar[data_3]",
    ]
    assert mod.coords["lag_shared_ar"] == [1, 2, 3]
    assert mod.coords["lag_individual_ar"] == [1, 2, 3]

    outputs = [mod.ssm["transition"], mod.ssm["design"], mod.ssm["selection"], mod.ssm["state_cov"]]
    T, Z, R, Q = pytensor.function(
        list(explicit_graph_inputs(outputs)),
        outputs,
    )(
        **{
            "params_shared_ar": np.array([0.9, 0.8, 0.7]),
            "params_individual_ar": np.full((3, 3), 0.5),
            "sigma_shared_ar": np.array(0.1),
            "sigma_individual_ar": np.array([0.05, 0.12, 0.22]),
        }
    )

    np.testing.assert_allclose(
        T,
        np.array(
            [
                [0.9, 0.8, 0.7, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.5, 0.5, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.5, 0.5, 0.5, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.5, 0.5, 0.5],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            ]
        ),
    )

    np.testing.assert_allclose(
        Z,
        np.array(
            [
                [1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            ]
        ),
    )

    np.testing.assert_allclose(
        R,
        np.array(
            [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ]
        ),
    )

    np.testing.assert_allclose(Q, np.diag([0.1, 0.05, 0.12, 0.22]) ** 2)



import numpy as np

from pytensor import tensor as pt

from pymc_extras.statespace.models.structural.core import Component
from pymc_extras.statespace.models.structural.utils import _frequency_transition_block


class TimeSeasonality(Component):
    r"""
    Seasonal component, modeled in the time domain

    Parameters
    ----------
    season_length: int
        The number of periods in a single seasonal cycle, e.g. 12 for monthly data with annual seasonal pattern, 7 for
        daily data with weekly seasonal pattern, etc. It must be greater than one.

    duration: int, default 1
        Number of time steps for each seasonal period.
        This determines how long each seasonal period is held constant before moving to the next.

    innovations: bool, default True
        Whether to include stochastic innovations in the strength of the seasonal effect

    name: str, default None
        A name for this seasonal component. Used to label dimensions and coordinates. Useful when multiple seasonal
        components are included in the same model. Default is ``f"Seasonal[s={season_length}, d={duration}]"``

    state_names: list of str, default None
        List of strings for seasonal effect labels. If provided, it must be of length ``season_length`` times ``duration``.
        An example would be ``state_names = ['Mon', 'Tue', 'Wed', 'Thur', 'Fri', 'Sat', 'Sun']`` when data is daily with a weekly
        seasonal pattern (``season_length = 7``).

        If None and ``duration = 1``, states will be named as ``[State_0, ..., State_s-1]`` (here s is ``season_length``).
        If None and ``duration > 1``, states will be named as ``[State_0_0, ..., State_s-1_d-1]`` (here d is ``duration``).

    remove_first_state: bool, default True
        If True, the first state will be removed from the model. This is done because there are only ``season_length-1`` degrees of
        freedom in the seasonal component, and one state is not identified. If False, the first state will be
        included in the model, but it will not be identified -- you will need to handle this in the priors (e.g. with
        ZeroSumNormal).

    observed_state_names: list[str] | None, default None
        List of strings for observed state labels. If None, defaults to ["data"].

    share_states: bool, default False
        Whether latent states are shared across the observed states. If True, there will be only one set of latent
        states, which are observed by all observed states. If False, each observed state has its own set of
        latent states. This argument has no effect if `k_endog` is 1.

    Notes
    -----
    A seasonal effect is any pattern that repeats at fixed intervals. There are several ways to model such effects;
    here, we present two models that are straightforward extensions of those described in [1].

    **First model** (``remove_first_state=True``)

    In this model, the state vector is defined as:

    .. math::
        \alpha_t :=(\gamma_t, \ldots, \gamma_{t-d(s-1)+1}), \quad t \ge 0.

    This vector has length :math:`d(s-1)`, where:

    - :math:`s` is the ``seasonal_length`` parameter, and
    - :math:`d` is the ``duration`` parameter.

    The components of the initial vector :math:`\alpha_{0}` are given by

    .. math::
        \gamma_{-l} := \tilde{\gamma}_{k_l}, \quad \text{where} \quad k_l := \left\lfloor \frac{l}{d} \right\rfloor \bmod s \quad \text{and} \quad l=0,\ldots, d(s-1)-1.

    Here, the values

    .. math::
        \tilde{\gamma}_{0}, \ldots, \tilde{\gamma}_{s-2},

    represent the initial seasonal states. The transition matrix of this model is the :math:`d(s-1) \times d(s-1)` matrix

    .. math::
        \begin{bmatrix}
            -\mathbf{1}_d & -\mathbf{1}_d & \cdots & -\mathbf{1}_d & -\mathbf{1}_d \\
            \mathbf{1}_d & \mathbf{0}_d & \cdots & \mathbf{0}_d & \mathbf{0}_d \\
            \mathbf{0}_d & \mathbf{1}_d & \cdots & \mathbf{0}_d & \mathbf{0}_d \\
            \vdots & \vdots & \ddots & \vdots \\
            \mathbf{0}_d & \mathbf{0}_d & \cdots & \mathbf{1}_d & \mathbf{0}_d
        \end{bmatrix}

    where :math:`\mathbf{1}_d` and  :math:`\mathbf{0}_d` denote the :math:`d \times d` identity and null matrices, respectively.

    **Second model** (``remove_first_state=False``)

    In contrast, the state vector in the second model is defined as:

    .. math::
        \alpha_t=(\gamma_t, \ldots, \gamma_{t-ds+1}), \quad t \ge 0.

    This vector has length :math:`ds`. The components of the initial state vector :math:`\alpha_{0}` are defined similarly:

    .. math::
        \gamma_{-l} := \tilde{\gamma}_{k_l}, \quad \text{where} \quad k_l := \left\lfloor \frac{l}{d} \right\rfloor \bmod s \quad \text{and} \quad l=0,\ldots, ds-1.

    In this case, the initial seasonal states :math:`\tilde{\gamma}_{0}, \ldots, \tilde{\gamma}_{s-1}` are required to satisfy the following condition:

    .. math::
        \sum_{i=0}^{s-1} \tilde{\gamma}_{i} = 0.

    The transition matrix of this model is the following :math:`ds \times ds` circulant matrix:

    .. math::
        \begin{bmatrix}
            0 & 1 & 0 & \cdots & 0 \\
            0 & 0 & 1 & \cdots & 0 \\
            \vdots & \vdots & \ddots & \ddots & \vdots \\
            0 & 0 & \cdots & 0 & 1 \\
            1 & 0 & \cdots & 0 & 0
        \end{bmatrix}

    To give interpretation to the :math:`\gamma` terms, it is helpful to work through the algebra for a simple
    example. Let :math:`s=4`, :math:`d=1`, ``remove_first_state=True``, and omit the shock term. Then, we have
    :math:`\gamma_{-i} = \tilde{\gamma}_{-i}`, for :math:`i=-2,\ldots, 0` and the value of the seasonal component
    for the first 5 timesteps will be:

    .. math::
        \begin{align}
            \gamma_1 &= -\gamma_0 - \gamma_{-1} - \gamma_{-2} \\
             \gamma_2 &= -\gamma_1 - \gamma_0 - \gamma_{-1} \\
                       &= -(-\gamma_0 - \gamma_{-1} - \gamma_{-2}) - \gamma_0 - \gamma_{-1}  \\
                       &= (\gamma_0 - \gamma_0 )+ (\gamma_{-1} - \gamma_{-1}) + \gamma_{-2} \\
                       &= \gamma_{-2} \\
              \gamma_3 &= -\gamma_2 - \gamma_1 - \gamma_0  \\
                       &= -\gamma_{-2} - (-\gamma_0 - \gamma_{-1} - \gamma_{-2}) - \gamma_0 \\
                       &=  (\gamma_{-2} - \gamma_{-2}) + \gamma_{-1} + (\gamma_0 - \gamma_0) \\
                       &= \gamma_{-1} \\
              \gamma_4 &= -\gamma_3 - \gamma_2 - \gamma_1 \\
                       &= -\gamma_{-1} - \gamma_{-2} -(-\gamma_0 - \gamma_{-1} - \gamma_{-2}) \\
                       &= (\gamma_{-2} - \gamma_{-2}) + (\gamma_{-1} - \gamma_{-1}) + \gamma_0 \\
                       &= \gamma_0 \\
              \gamma_5 &= -\gamma_4 - \gamma_3 - \gamma_2 \\
                       &= -\gamma_0 - \gamma_{-1} - \gamma_{-2} \\
                       &= \gamma_1
        \end{align}

    This exercise shows that, given a list ``initial_conditions`` of length ``s-1``, the effects of this model will be:

        - Period 1: ``-sum(initial_conditions)``
        - Period 2: ``initial_conditions[-1]``
        - Period 3: ``initial_conditions[-2]``
        - ...
        - Period s: ``initial_conditions[0]``
        - Period s+1: ``-sum(initial_condition)``

    And so on. So for interpretation, the ``season_length - 1`` initial states are, when reversed, the coefficients
    associated with ``state_names[1:]``.

    In the next example, we set :math:`s=2`, :math:`d=2`, ``remove_first_state=True``, and omit the shock term.
    By definition, the initial vector :math:`\alpha_{0}` is

    .. math::
        \alpha_0=(\tilde{\gamma}_{0}, \tilde{\gamma}_{0}, \tilde{\gamma}_{-1}, \tilde{\gamma}_{-1})

    and the transition matrix is

    .. math::
        \begin{bmatrix}
            -1 &  0 & -1 &  0 \\
             0 & -1 &  0 & -1 \\
             1 &  0 &  0 &  0 \\
             0 &  1 &  0 &  0 \\
        \end{bmatrix}

    It is easy to verify that:

    .. math::
        \begin{align}
            \gamma_1 &= -\tilde{\gamma}_0 - \tilde{\gamma}_{-1}\\
            \gamma_2 &= -(-\tilde{\gamma}_0 - \tilde{\gamma}_{-1})-\tilde{\gamma}_0\\
                     &= \tilde{\gamma}_{-1}\\
            \gamma_3 &= -\tilde{\gamma}_{-1} +(\tilde{\gamma}_0 + \tilde{\gamma}_{-1})\\
                     &= \tilde{\gamma}_{0}\\
            \gamma_4 &= -\tilde{\gamma}_0 - \tilde{\gamma}_{-1}.\\
        \end{align}

    .. warning::
        Although the ``state_names`` argument expects a list of length ``season_length`` times ``duration``,
        only ``state_names[duration:]`` will be saved as model dimensions, since the first coefficient is not identified
        (it is defined as :math:`-\sum_{i=1}^{s-1} \tilde{\gamma}_{-i}`).

    Examples
    --------
    Estimate monthly with a model with a gaussian random walk trend and monthly seasonality:

    .. code:: python

        from pymc_extras.statespace import structural as st
        import pymc as pm
        import pytensor.tensor as pt
        import pandas as pd

        # Get month names
        state_names = pd.date_range('1900-01-01', '1900-12-31', freq='MS').month_name().tolist()

        # Build the structural model
        grw = st.LevelTrendComponent(order=1, innovations_order=1)
        annual_season = st.TimeSeasonality(
            season_length=12, name="annual", state_names=state_names, innovations=False
        )
        ss_mod = (grw + annual_season).build()

        with pm.Model(coords=ss_mod.coords) as model:
            P0 = pm.Deterministic('P0', pt.eye(ss_mod.k_states) * 10, dims=ss_mod.param_dims['P0'])

            initial_level_trend = pm.Deterministic(
                "initial_level_trend", pt.zeros(1), dims=ss_mod.param_dims["initial_level_trend"]
            )
            sigma_level_trend = pm.HalfNormal(
                "sigma_level_trend", sigma=1e-6, dims=ss_mod.param_dims["sigma_level_trend"]
            )
            params_annual = pm.Normal("params_annual", sigma=1e-2, dims=ss_mod.param_dims["params_annual"])

            ss_mod.build_statespace_graph(data)
            idata = pm.sample(
                nuts_sampler="nutpie", nuts_sampler_kwargs={"backend": "JAX", "gradient_backend": "JAX"}
            )

    References
    ----------
    .. [1] Durbin, James, and Siem Jan Koopman. 2012.
        Time Series Analysis by State Space Methods: Second Edition.
        Oxford University Press.
    """

    def __init__(
        self,
        season_length: int,
        duration: int = 1,
        innovations: bool = True,
        name: str | None = None,
        state_names: list | None = None,
        remove_first_state: bool = True,
        observed_state_names: list[str] | None = None,
        share_states: bool = False,
    ):
        if observed_state_names is None:
            observed_state_names = ["data"]

        if season_length <= 1 or not isinstance(season_length, int):
            raise ValueError(
                f"season_length must be an integer greater than 1, got {season_length}"
            )
        if duration <= 0 or not isinstance(duration, int):
            raise ValueError(f"duration must be a positive integer, got {duration}")
        if name is None:
            name = f"Seasonal[s={season_length}, d={duration}]"
        if state_names is None:
            if duration > 1:
                state_names = [
                    f"{name}_{i}_{j}" for i in range(season_length) for j in range(duration)
                ]
            else:
                state_names = [f"{name}_{i}" for i in range(season_length)]
        else:
            if len(state_names) != season_length * duration:
                raise ValueError(
                    f"state_names must be a list of length season_length*duration, got {len(state_names)}"
                )
            state_names = state_names.copy()

        self.share_states = share_states
        self.innovations = innovations
        self.duration = duration
        self.remove_first_state = remove_first_state
        self.season_length = season_length

        if self.remove_first_state:
            # In traditional models, the first state isn't identified, so we can help out the user by automatically
            # discarding it.
            # TODO: Can this be stashed and reconstructed automatically somehow?
            state_names = state_names[duration:]

        self.provided_state_names = state_names

        k_states = (season_length - int(self.remove_first_state)) * duration
        k_endog = len(observed_state_names)
        k_posdef = int(innovations)

        super().__init__(
            name=name,
            k_endog=k_endog,
            k_states=k_states if share_states else k_states * k_endog,
            k_posdef=k_posdef if share_states else k_posdef * k_endog,
            observed_state_names=observed_state_names,
            measurement_error=False,
            combine_hidden_states=True,
            obs_state_idxs=np.tile(
                np.array([1.0] + [0.0] * (k_states - 1)), 1 if share_states else k_endog
            ),
            share_states=share_states,
        )

    def populate_component_properties(self):
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective

        if self.share_states:
            self.state_names = [
                f"{state_name}[{self.name}_shared]" for state_name in self.provided_state_names
            ]
        else:
            self.state_names = [
                f"{state_name}[{endog_name}]"
                for endog_name in self.observed_state_names
                for state_name in self.provided_state_names
            ]

        self.param_names = [f"params_{self.name}"]

        self.param_info = {
            f"params_{self.name}": {
                "shape": (k_states,) if k_endog == 1 else (k_endog, k_states),
                "constraints": None,
                "dims": (f"state_{self.name}",)
                if k_endog_effective == 1
                else (f"endog_{self.name}", f"state_{self.name}"),
            }
        }

        self.param_dims = {
            f"params_{self.name}": (f"state_{self.name}",)
            if k_endog_effective == 1
            else (f"endog_{self.name}", f"state_{self.name}")
        }

        self.coords = (
            {f"state_{self.name}": self.provided_state_names}
            if k_endog_effective == 1
            else {
                f"endog_{self.name}": self.observed_state_names,
                f"state_{self.name}": self.provided_state_names,
            }
        )

        if self.innovations:
            self.param_names += [f"sigma_{self.name}"]
            self.param_info[f"sigma_{self.name}"] = {
                "shape": () if k_endog_effective == 1 else (k_endog,),
                "constraints": "Positive",
                "dims": None if k_endog_effective == 1 else (f"endog_{self.name}",),
            }
            if self.share_states:
                self.shock_names = [f"{self.name}[shared]"]
            else:
                self.shock_names = [f"{self.name}[{name}]" for name in self.observed_state_names]

            if k_endog > 1:
                self.param_dims[f"sigma_{self.name}"] = (f"endog_{self.name}",)

    def make_symbolic_graph(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog
        k_states = self.k_states // k_endog_effective
        duration = self.duration

        k_unique_states = k_states // duration
        k_posdef = self.k_posdef // k_endog_effective

        if self.remove_first_state:
            # In this case, parameters are normalized to sum to zero, so the current state is the negative sum of
            # all previous states.
            zero_d = pt.zeros((self.duration, self.duration))
            id_d = pt.eye(self.duration)

            row_blocks = []

            # First row: all -1_d blocks
            first_row = [-id_d for _ in range(self.season_length - 1)]
            row_blocks.append(pt.concatenate(first_row, axis=1))

            # Rows 2 to season_length-1: shifted identity blocks
            for i in range(self.season_length - 2):
                row = []
                for j in range(self.season_length - 1):
                    if j == i:
                        row.append(id_d)
                    else:
                        row.append(zero_d)
                row_blocks.append(pt.concatenate(row, axis=1))

            # Stack blocks
            T = pt.concatenate(row_blocks, axis=0)
        else:
            # In this case we assume the user to be responsible for ensuring the states sum to zero, so T is just a
            # circulant matrix that cycles between the states.
            T = pt.eye(k_states, k=1)
            T = pt.set_subtensor(T[-1, 0], 1)

        self.ssm["transition", :, :] = pt.linalg.block_diag(*[T for _ in range(k_endog_effective)])

        Z = pt.zeros((1, k_states))[0, 0].set(1)
        self.ssm["design", :, :] = pt.linalg.block_diag(*[Z for _ in range(k_endog_effective)])

        initial_states = self.make_and_register_variable(
            f"params_{self.name}",
            shape=(k_unique_states,)
            if k_endog_effective == 1
            else (k_endog_effective, k_unique_states),
        )
        if k_endog_effective == 1:
            self.ssm["initial_state", :] = pt.extra_ops.repeat(initial_states, duration, axis=0)
        else:
            self.ssm["initial_state", :] = pt.extra_ops.repeat(
                initial_states, duration, axis=1
            ).ravel()

        if self.innovations:
            R = pt.zeros((k_states, k_posdef))[0, 0].set(1.0)
            self.ssm["selection", :, :] = pt.join(0, *[R for _ in range(k_endog_effective)])
            season_sigma = self.make_and_register_variable(
                f"sigma_{self.name}", shape=() if k_endog_effective == 1 else (k_endog_effective,)
            )
            cov_idx = ("state_cov", *np.diag_indices(k_posdef * k_endog_effective))
            self.ssm[cov_idx] = season_sigma**2


class FrequencySeasonality(Component):
    r"""
    Seasonal component, modeled in the frequency domain

    Parameters
    ----------
    season_length: float
        The number of periods in a single seasonal cycle, e.g. 12 for monthly data with annual seasonal pattern, 7 for
        daily data with weekly seasonal pattern, etc. Non-integer seasonal_length is also permitted, for example
        365.2422 days in a (solar) year.

    n: int
        Number of fourier features to include in the seasonal component. Default is ``season_length // 2``, which
        is the maximum possible. A smaller number can be used for a more wave-like seasonal pattern.

    name: str, default None
        A name for this seasonal component. Used to label dimensions and coordinates. Useful when multiple seasonal
        components are included in the same model. Default is ``f"Seasonal[s={season_length}, n={n}]"``

    innovations: bool, default True
        Whether to include stochastic innovations in the strength of the seasonal effect

    observed_state_names: list[str] | None, default None
        List of strings for observed state labels. If None, defaults to ["data"].

    share_states: bool, default False
        Whether latent states are shared across the observed states. If True, there will be only one set of latent
        states, which are observed by all observed states. If False, each observed state has its own set of
        latent states. This argument has no effect if `k_endog` is 1.

    Notes
    -----
    A seasonal effect is any pattern that repeats every fixed interval. Although there are many possible ways to
    model seasonal effects, the implementation used here is the one described by [1] as the "canonical" frequency domain
    representation. The seasonal component can be expressed:

    .. math::
        \begin{align}
            \gamma_t &= \sum_{j=1}^{2n} \gamma_{j,t} \\
            \gamma_{j, t+1} &= \gamma_{j,t} \cos \lambda_j + \gamma_{j,t}^\star \sin \lambda_j + \omega_{j, t} \\
            \gamma_{j, t}^\star &= -\gamma_{j,t} \sin \lambda_j + \gamma_{j,t}^\star \cos \lambda_j + \omega_{j,t}^\star
            \lambda_j &= \frac{2\pi j}{s}
        \end{align}

    Where :math:`s` is the ``seasonal_length``.

    Unlike a ``TimeSeasonality`` component, a ``FrequencySeasonality`` component does not require integer season
    length. In addition, for long seasonal periods, it is possible to obtain a more compact state space representation
    by choosing ``n << s // 2``. Using ``TimeSeasonality``, an annual seasonal pattern in daily data requires 364
    states, whereas ``FrequencySeasonality`` always requires ``2 * n`` states, regardless of the ``seasonal_length``.
    The price of this compactness is less representational power. At ``n = 1``, the seasonal pattern will be a pure
    sine wave. At ``n = s // 2``, any arbitrary pattern can be represented.

    One cost of the added flexibility of ``FrequencySeasonality`` is reduced interpretability. States of this model are
    coefficients :math:`\gamma_1, \gamma^\star_1, \gamma_2, \gamma_2^\star ..., \gamma_n, \gamma^\star_n` associated
    with different frequencies in the fourier representation of the seasonal pattern. As a result, it is not possible
    to isolate and identify a "Monday" effect, for instance.
    """

    def __init__(
        self,
        season_length: int,
        n: int | None = None,
        name: str | None = None,
        innovations: bool = True,
        observed_state_names: list[str] | None = None,
        share_states: bool = False,
    ):
        if observed_state_names is None:
            observed_state_names = ["data"]

        self.share_states = share_states
        k_endog = len(observed_state_names)

        if n is None:
            n = int(season_length / 2)
        if name is None:
            name = f"Frequency[s={season_length}, n={n}]"

        k_states = n * 2
        self.n = n
        self.season_length = season_length
        self.innovations = innovations

        # If the model is completely saturated (n = s // 2), the last state will not be identified, so it shouldn't
        # get a parameter assigned to it and should just be fixed to zero.
        # Test this way (rather than n == s // 2) to catch cases when n is non-integer.
        self.last_state_not_identified = (self.season_length / self.n) == 2.0
        self.n_coefs = k_states - int(self.last_state_not_identified)

        obs_state_idx = np.zeros(k_states)
        obs_state_idx[slice(0, k_states, 2)] = 1
        obs_state_idx = np.tile(obs_state_idx, 1 if share_states else k_endog)

        super().__init__(
            name=name,
            k_endog=k_endog,
            k_states=k_states if share_states else k_states * k_endog,
            k_posdef=k_states * int(self.innovations)
            if share_states
            else k_states * int(self.innovations) * k_endog,
            share_states=share_states,
            observed_state_names=observed_state_names,
            measurement_error=False,
            combine_hidden_states=True,
            obs_state_idxs=obs_state_idx,
        )

    def make_symbolic_graph(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective
        k_posdef = self.k_posdef // k_endog_effective
        n_coefs = self.n_coefs

        Z = pt.zeros((1, k_states))[0, slice(0, k_states, 2)].set(1.0)

        self.ssm["design", :, :] = pt.linalg.block_diag(*[Z for _ in range(k_endog_effective)])

        init_state = self.make_and_register_variable(
            f"params_{self.name}", shape=(n_coefs,) if k_endog == 1 else (k_endog, n_coefs)
        )

        init_state_idx = np.concatenate(
            [
                np.arange(k_states * i, (i + 1) * k_states, dtype=int)[:n_coefs]
                for i in range(k_endog_effective)
            ],
            axis=0,
        )

        self.ssm["initial_state", init_state_idx] = init_state.ravel()

        T_mats = [_frequency_transition_block(self.season_length, j + 1) for j in range(self.n)]
        T = pt.linalg.block_diag(*T_mats)
        self.ssm["transition", :, :] = pt.linalg.block_diag(*[T for _ in range(k_endog_effective)])

        if self.innovations:
            sigma_season = self.make_and_register_variable(
                f"sigma_{self.name}", shape=() if k_endog_effective == 1 else (k_endog_effective,)
            )
            self.ssm["selection", :, :] = pt.eye(self.k_states)
            self.ssm["state_cov", :, :] = pt.eye(self.k_posdef) * pt.repeat(
                sigma_season**2, k_posdef
            )

    def populate_component_properties(self):
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog
        n_coefs = self.n_coefs

        base_names = [f"{f}_{i}_{self.name}" for i in range(self.n) for f in ["Cos", "Sin"]]

        if self.share_states:
            self.state_names = [f"{name}[shared]" for name in base_names]
        else:
            self.state_names = [
                f"{name}[{obs_state_name}]"
                for obs_state_name in self.observed_state_names
                for name in base_names
            ]

        # Trim state names if the model is saturated
        param_state_names = base_names[:n_coefs]

        self.param_names = [f"params_{self.name}"]
        self.param_dims = {
            f"params_{self.name}": (f"state_{self.name}",)
            if k_endog_effective == 1
            else (f"endog_{self.name}", f"state_{self.name}")
        }
        self.param_info = {
            f"params_{self.name}": {
                "shape": (n_coefs,) if k_endog_effective == 1 else (k_endog_effective, n_coefs),
                "constraints": None,
                "dims": (f"state_{self.name}",)
                if k_endog_effective == 1
                else (f"endog_{self.name}", f"state_{self.name}"),
            }
        }

        self.coords = (
            {f"state_{self.name}": param_state_names}
            if k_endog == 1
            else {
                f"endog_{self.name}": self.observed_state_names,
                f"state_{self.name}": param_state_names,
            }
        )

        if self.innovations:
            self.param_names += [f"sigma_{self.name}"]
            self.shock_names = self.state_names.copy()
            self.param_info[f"sigma_{self.name}"] = {
                "shape": () if k_endog_effective == 1 else (k_endog_effective, n_coefs),
                "constraints": "Positive",
                "dims": None if k_endog_effective == 1 else (f"endog_{self.name}",),
            }
            if k_endog_effective > 1:
                self.param_dims[f"sigma_{self.name}"] = (f"endog_{self.name}",)



import arviz as az
import numpy as np
import pymc
import pytensor
import pytensor.tensor as pt
import xarray

from better_optimize import minimize
from better_optimize.constants import minimize_method
from pymc import DictToArrayBijection, Model, join_nonshared_inputs
from pymc.backends.arviz import (
    PointFunc,
    apply_function_over_dataset,
    coords_and_dims_for_inferencedata,
)
from pymc.util import RandomSeed, get_default_varnames
from pytensor.tensor.variable import TensorVariable

from pymc_extras.inference.laplace_approx.laplace import unstack_laplace_draws
from pymc_extras.inference.laplace_approx.scipy_interface import (
    _compile_functions_for_scipy_optimize,
)


def fit_dadvi(
    model: Model | None = None,
    n_fixed_draws: int = 30,
    random_seed: RandomSeed = None,
    n_draws: int = 1000,
    keep_untransformed: bool = False,
    optimizer_method: minimize_method = "trust-ncg",
    use_grad: bool = True,
    use_hessp: bool = True,
    use_hess: bool = False,
    **minimize_kwargs,
) -> az.InferenceData:
    """
    Does inference using deterministic ADVI (automatic differentiation
    variational inference), DADVI for short.

    For full details see the paper cited in the references:
    https://www.jmlr.org/papers/v25/23-1015.html

    Parameters
    ----------
    model : pm.Model
        The PyMC model to be fit. If None, the current model context is used.

    n_fixed_draws : int
        The number of fixed draws to use for the optimisation. More
        draws will result in more accurate estimates, but also
        increase inference time. Usually, the default of 30 is a good
        tradeoff.between speed and accuracy.

    random_seed: int
        The random seed to use for the fixed draws. Running the optimisation
        twice with the same seed should arrive at the same result.

    n_draws: int
        The number of draws to return from the variational approximation.

    keep_untransformed: bool
        Whether or not to keep the unconstrained variables (such as
        logs of positive-constrained parameters) in the output.

    optimizer_method: str
        Which optimization method to use. The function calls
        ``scipy.optimize.minimize``, so any of the methods there can
        be used. The default is trust-ncg, which uses second-order
        information and is generally very reliable. Other methods such
        as L-BFGS-B might be faster but potentially more brittle and
        may not converge exactly to the optimum.

    minimize_kwargs:
        Additional keyword arguments to pass to the
        ``scipy.optimize.minimize`` function. See the documentation of
        that function for details.

    use_grad:
        If True, pass the gradient function to
        `scipy.optimize.minimize` (where it is referred to as `jac`).

    use_hessp:
        If True, pass the hessian vector product to `scipy.optimize.minimize`.

    use_hess:
        If True, pass the hessian to `scipy.optimize.minimize`. Note that
        this is generally not recommended since its computation can be slow
        and memory-intensive if there are many parameters.

    Returns
    -------
    :class:`~arviz.InferenceData`
        The inference data containing the results of the DADVI algorithm.

    References
    ----------
    Giordano, R., Ingram, M., & Broderick, T. (2024). Black Box
    Variational Inference with a Deterministic Objective: Faster, More
    Accurate, and Even More Black Box. Journal of Machine Learning
    Research, 25(18), 1–39.
    """

    model = pymc.modelcontext(model) if model is None else model

    initial_point_dict = model.initial_point()
    n_params = DictToArrayBijection.map(initial_point_dict).data.shape[0]

    var_params, objective = create_dadvi_graph(
        model,
        n_fixed_draws=n_fixed_draws,
        random_seed=random_seed,
        n_params=n_params,
    )

    f_fused, f_hessp = _compile_functions_for_scipy_optimize(
        objective,
        [var_params],
        compute_grad=use_grad,
        compute_hessp=use_hessp,
        compute_hess=use_hess,
    )

    derivative_kwargs = {}

    if use_grad:
        derivative_kwargs["jac"] = True
    if use_hessp:
        derivative_kwargs["hessp"] = f_hessp
    if use_hess:
        derivative_kwargs["hess"] = True

    result = minimize(
        f_fused,
        np.zeros(2 * n_params),
        method=optimizer_method,
        **derivative_kwargs,
        **minimize_kwargs,
    )

    opt_var_params = result.x
    opt_means, opt_log_sds = np.split(opt_var_params, 2)

    # Make the draws:
    generator = np.random.default_rng(seed=random_seed)
    draws_raw = generator.standard_normal(size=(n_draws, n_params))

    draws = opt_means + draws_raw * np.exp(opt_log_sds)
    draws_arviz = unstack_laplace_draws(draws, model, chains=1, draws=n_draws)

    transformed_draws = transform_draws(draws_arviz, model, keep_untransformed=keep_untransformed)

    return transformed_draws


def create_dadvi_graph(
    model: Model,
    n_params: int,
    n_fixed_draws: int = 30,
    random_seed: RandomSeed = None,
) -> tuple[TensorVariable, TensorVariable]:
    """
    Sets up the DADVI graph in pytensor and returns it.

    Parameters
    ----------
    model : pm.Model
        The PyMC model to be fit.

    n_params: int
        The total number of parameters in the model.

    n_fixed_draws : int
        The number of fixed draws to use.

    random_seed: int
        The random seed to use for the fixed draws.

    Returns
    -------
    Tuple[TensorVariable, TensorVariable]
        A tuple whose first element contains the variational parameters,
        and whose second contains the DADVI objective.
    """

    # Make the fixed draws
    generator = np.random.default_rng(seed=random_seed)
    draws = generator.standard_normal(size=(n_fixed_draws, n_params))

    inputs = model.continuous_value_vars + model.discrete_value_vars
    initial_point_dict = model.initial_point()
    logp = model.logp()

    # Graph in terms of a flat input
    [logp], flat_input = join_nonshared_inputs(
        point=initial_point_dict, outputs=[logp], inputs=inputs
    )

    var_params = pt.vector(name="eta", shape=(2 * n_params,))

    means, log_sds = pt.split(var_params, axis=0, splits_size=[n_params, n_params], n_splits=2)

    draw_matrix = pt.constant(draws)
    samples = means + pt.exp(log_sds) * draw_matrix

    logp_vectorized_draws = pytensor.graph.vectorize_graph(logp, replace={flat_input: samples})

    mean_log_density = pt.mean(logp_vectorized_draws)
    entropy = pt.sum(log_sds)

    objective = -mean_log_density - entropy

    return var_params, objective


def transform_draws(
    unstacked_draws: xarray.Dataset,
    model: Model,
    keep_untransformed: bool = False,
):
    """
    Transforms the unconstrained draws back into the constrained space.

    Parameters
    ----------
    unstacked_draws : xarray.Dataset
        The draws to constrain back into the original space.

    model : Model
        The PyMC model the variables were derived from.

    n_draws: int
        The number of draws to return from the variational approximation.

    keep_untransformed: bool
        Whether or not to keep the unconstrained variables in the output.

    Returns
    -------
    :class:`~arviz.InferenceData`
        Draws from the original constrained parameters.
    """

    filtered_var_names = model.unobserved_value_vars
    vars_to_sample = list(
        get_default_varnames(filtered_var_names, include_transformed=keep_untransformed)
    )
    fn = pytensor.function(model.value_vars, vars_to_sample)
    point_func = PointFunc(fn)

    coords, dims = coords_and_dims_for_inferencedata(model)

    transformed_result = apply_function_over_dataset(
        point_func,
        unstacked_draws,
        output_var_names=[x.name for x in vars_to_sample],
        coords=coords,
        dims=dims,
    )

    return transformed_result
